#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType


# pylint: disable=too-many-instance-attributes
class GXDLMSG3PlcMacLayerCounters(GXDLMSObject, IGXDLMSBase):

    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSData
    """

    def __init__(self, ln="0.0.29.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.G3_PLC_MAC_LAYER_COUNTERS, ln, sn)
        self.version = 1
        self.txDataPacketCount = 0
        self.rxDataPacketCount = 0
        self.txCmdPacketCount = 0
        self.rxCmdPacketCount = 0
        self.csmaFailCount = 0
        self.csmaNoAckCount = 0
        self.badCrcCount = 0
        self.txDataBroadcastCount = 0
        self.rxDataBroadcastCount = 0

    def getValues(self):
        return [
            self.logicalName,
            self.txDataPacketCount,
            self.rxDataPacketCount,
            self.txCmdPacketCount,
            self.rxCmdPacketCount,
            self.csmaFailCount,
            self.csmaNoAckCount,
            self.badCrcCount,
            self.txDataBroadcastCount,
            self.rxDataBroadcastCount,
        ]

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        # LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        if all or self.canRead(2):
            attributes.append(2)
        if all or self.canRead(3):
            attributes.append(3)
        if all or self.canRead(4):
            attributes.append(4)
        if all or self.canRead(5):
            attributes.append(5)
        if all or self.canRead(6):
            attributes.append(6)
        if all or self.canRead(7):
            attributes.append(7)
        if all or self.canRead(8):
            attributes.append(8)
        if all or self.canRead(9):
            attributes.append(9)
        if all or self.canRead(10):
            attributes.append(10)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 10

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 1

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        elif index in range(2, 11):
            # pylint: disable=super-with-arguments
            return DataType.UINT32
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.txDataPacketCount
        elif e.index == 3:
            ret = self.rxDataPacketCount
        elif e.index == 4:
            ret = self.txCmdPacketCount
        elif e.index == 5:
            ret = self.rxCmdPacketCount
        elif e.index == 6:
            ret = self.csmaFailCount
        elif e.index == 7:
            ret = self.csmaNoAckCount
        elif e.index == 8:
            ret = self.badCrcCount
        elif e.index == 9:
            ret = self.txDataBroadcastCount
        elif e.index == 10:
            ret = self.rxDataBroadcastCount
        else:
            ret = None
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.txDataPacketCount = int(e.value)
        elif e.index == 3:
            self.rxDataPacketCount = int(e.value)
        elif e.index == 4:
            self.txCmdPacketCount = int(e.value)
        elif e.index == 5:
            self.rxCmdPacketCount = int(e.value)
        elif e.index == 6:
            self.csmaFailCount = int(e.value)
        elif e.index == 7:
            self.csmaNoAckCount = int(e.value)
        elif e.index == 8:
            self.badCrcCount = int(e.value)
        elif e.index == 9:
            self.txDataBroadcastCount = int(e.value)
        elif e.index == 10:
            self.rxDataBroadcastCount = int(e.value)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.txDataPacketCount = reader.readElementContentAsInt("TxDataPacketCount")
        self.rxDataPacketCount = reader.readElementContentAsInt("RxDataPacketCount")
        self.txCmdPacketCount = reader.readElementContentAsInt("TxCmdPacketCount")
        self.rxCmdPacketCount = reader.readElementContentAsInt("RxCmdPacketCount")
        self.csmaFailCount = reader.readElementContentAsInt("CSMAFailCount")
        self.csmaNoAckCount = reader.readElementContentAsInt("CSMANoAckCount")
        self.badCrcCount = reader.readElementContentAsInt("BadCrcCount")
        self.txDataBroadcastCount = reader.readElementContentAsInt(
            "TxDataBroadcastCount"
        )
        self.rxDataBroadcastCount = reader.readElementContentAsInt(
            "RxDataBroadcastCount"
        )

    def save(self, writer):
        writer.writeElementString("TxDataPacketCount", self.txDataPacketCount)
        writer.writeElementString("RxDataPacketCount", self.rxDataPacketCount)
        writer.writeElementString("TxCmdPacketCount", self.txCmdPacketCount)
        writer.writeElementString("RxCmdPacketCount", self.rxCmdPacketCount)
        writer.writeElementString("CSMAFailCount", self.csmaFailCount)
        writer.writeElementString("CSMANoAckCount", self.csmaNoAckCount)
        writer.writeElementString("BadCrcCount", self.badCrcCount)
        writer.writeElementString("TxDataBroadcastCount", self.txDataBroadcastCount)
        writer.writeElementString("RxDataBroadcastCount", self.rxDataBroadcastCount)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType


# pylint: disable=too-many-instance-attributes
class GXDLMSTcpUdpSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSTcpUdpSetup
    """

    def __init__(self, ln="0.0.25.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        super(GXDLMSTcpUdpSetup, self).__init__(ObjectType.TCP_UDP_SETUP, ln, sn)
        self.port = 4059
        self.inactivityTimeout = 180
        self.maximumSegmentSize = 576
        self.ipReference = None
        self.maximumSimultaneousConnections = 0

    def getValues(self):
        return [
            self.logicalName,
            self.port,
            self.ipReference,
            self.maximumSegmentSize,
            self.maximumSimultaneousConnections,
            self.inactivityTimeout,
        ]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Port
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  IPReference
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  MaximumSegmentSize
        if all_ or not self.isRead(4):
            attributes.append(4)
        #  MaximumSimultaneousConnections
        if all_ or not self.isRead(5):
            attributes.append(5)
        #  InactivityTimeout
        if all_ or not self.isRead(6):
            attributes.append(6)
        return attributes

    def getAttributeCount(self):
        return 6

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.UINT16
        elif index == 3:
            ret = DataType.OCTET_STRING
        elif index == 4:
            ret = DataType.UINT16
        elif index == 5:
            ret = DataType.UINT8
        elif index == 6:
            ret = DataType.UINT16
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.port
        elif e.index == 3:
            ret = _GXCommon.logicalNameToBytes(self.ipReference)
        elif e.index == 4:
            ret = self.maximumSegmentSize
        elif e.index == 5:
            ret = self.maximumSimultaneousConnections
        elif e.index == 6:
            ret = self.inactivityTimeout
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if e.value is None:
                self.port = 4059
            else:
                self.port = e.value
        elif e.index == 3:
            if e.value is None:
                self.ipReference = None
            else:
                if isinstance(e.value, bytearray):
                    self.ipReference = _GXCommon.toLogicalName(e.value)
                else:
                    self.ipReference(e.value)
        elif e.index == 4:
            if e.value is None:
                self.maximumSegmentSize = 576
            else:
                self.maximumSegmentSize = e.value
        elif e.index == 5:
            if e.value is None:
                self.maximumSimultaneousConnections = 1
            else:
                self.maximumSimultaneousConnections = e.value
        elif e.index == 6:
            if e.value is None:
                self.inactivityTimeout = 180
            else:
                self.inactivityTimeout = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.port = reader.readElementContentAsInt("Port")
        self.ipReference = reader.readElementContentAsString("IPReference")
        self.maximumSegmentSize = reader.readElementContentAsInt("MaximumSegmentSize")
        self.maximumSimultaneousConnections = reader.readElementContentAsInt(
            "MaximumSimultaneousConnections"
        )
        self.inactivityTimeout = reader.readElementContentAsInt("InactivityTimeout")

    def save(self, writer):
        writer.writeElementString("Port", self.port)
        writer.writeElementString("IPReference", self.ipReference)
        writer.writeElementString("MaximumSegmentSize", self.maximumSegmentSize)
        writer.writeElementString(
            "MaximumSimultaneousConnections", self.maximumSimultaneousConnections
        )
        writer.writeElementString("InactivityTimeout", self.inactivityTimeout)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import xml.etree.cElementTree as ET
from xml.dom import minidom
from .GXDLMSObject import GXDLMSObject
from ..enums import ObjectType
from .GXXmlWriter import GXXmlWriter
from .GXXmlReader import GXXmlReader
from ..GXDLMSConverter import GXDLMSConverter
from ..enums import AccessMode
from ..enums import AccessMode3
from ..enums import MethodAccessMode
from ..enums import MethodAccessMode3


class GXDLMSObjectCollection(list):
    """
    Collection of DLMS objects.
    """

    #
    # Constructor.
    #
    # forParent: Parent object.
    #
    def __init__(self, forParent=None):
        # pylint: disable=super-with-arguments
        super(GXDLMSObjectCollection, self).__init__()
        self.parent = forParent

    def append(self, item):
        if not isinstance(item, GXDLMSObject):
            raise TypeError("item is not of type GXDLMSObject")
        # pylint: disable=super-with-arguments
        super(GXDLMSObjectCollection, self).append(item)
        item.parent = self

    def extend(self, items):
        # pylint: disable=super-with-arguments
        if not isinstance(items, GXDLMSObjectCollection):
            raise TypeError("items is not of type GXDLMSObjectCollection")
        for it in items:
            super(GXDLMSObjectCollection, self).append(it)
            it.parent = self

    def getObjects(self, type_):
        if isinstance(type_, (int, ObjectType)):
            type_ = [type_]
        items = GXDLMSObjectCollection()
        for it in self:
            if it.objectType in type_:
                items.append(it)
        return items

    def findByLN(self, type_, ln):
        for it in self:
            if (
                type_ in (ObjectType.NONE, it.objectType)
                and it.logicalName.strip() == ln
            ):
                return it
        return None

    def findBySN(self, sn):
        for it in self:
            if it.shortName == sn:
                return it
        return None

    def __str__(self):
        str_ = "["
        for it in self:
            if str_:
                str_ += ", "
            str_ += it
        str_ += "]"
        return str_

    @classmethod
    def load(cls, file_):
        # pylint: disable=import-outside-toplevel
        from .._GXObjectFactory import _GXObjectFactory

        obj = None
        target = None
        type_ = None
        reader = GXXmlReader(file_, GXDLMSObjectCollection())
        reader.getNext()
        while not reader.isEOF():
            if reader.isStartElement():
                target = reader.name()
                if "Objects".lower() == target.lower():
                    reader.getNext()
                elif target.find("GXDLMS") == 0:
                    try:
                        type_ = GXDLMSConverter.valueOfObjectType(target[6:])
                    except Exception:
                        # pylint:disable=raise-missing-from
                        raise ValueError("Invalid object type: " + target + ".")
                    obj = _GXObjectFactory.createObject(type_)
                    obj.version = 0
                    reader.objects.append(obj)
                    reader.getNext()
                elif "SN".lower() == target.lower():
                    obj.shortName = reader.readElementContentAsInt("SN")
                elif "LN".lower() == target.lower():
                    obj.logicalName = reader.readElementContentAsString("LN")
                elif "Description".lower() == target.lower():
                    obj.description = reader.readElementContentAsString("Description")
                elif "Version".lower() == target.lower():
                    obj.version = reader.readElementContentAsInt("Version")
                elif "Access".lower() == target.lower():
                    pos = 0
                    tmp = reader.readElementContentAsString("Access")
                    if tmp:
                        for it in tmp:
                            pos += 1
                            obj.setAccess(pos, (AccessMode)(int(it)))
                elif "Access3".lower() == target.lower():
                    tmp = reader.readElementContentAsString("Access3")
                    if tmp:
                        pos = 0
                        while pos < len(tmp) / 4:
                            obj.setAccess3(
                                1 + pos,
                                (AccessMode3)(
                                    int(tmp[4 * pos : 4 * pos + 4], 16) & ~0x8000
                                ),
                            )
                            pos += 1
                elif "MethodAccess".lower() == target.lower():
                    pos = 0
                    tmp = reader.readElementContentAsString("MethodAccess")
                    if tmp:
                        for it in tmp:
                            pos += 1
                            obj.setMethodAccess(pos, (MethodAccessMode)(int(it)))
                elif "MethodAccess3".lower() == target.lower():
                    tmp = reader.readElementContentAsString("MethodAccess3")
                    if tmp:
                        pos = 0
                        while pos < len(tmp) / 4:
                            obj.setMethodAccess3(
                                1 + pos,
                                (MethodAccessMode3)(
                                    int(tmp[4 * pos : 4 * pos + 4], 16) & ~0x8000
                                ),
                            )
                            pos += 1
                else:
                    # pylint:disable=broad-except
                    try:
                        obj.load(reader)
                    except Exception as ex:
                        print("Failed to load object " + str(obj) + " " + str(ex))
                    obj = None
            else:
                reader.read()
        for obj in reader.objects:
            obj.postLoad(reader)
        return reader.objects

    def save(self, name, settings=None):
        from .GXDLMSAssociationLogicalName import GXDLMSAssociationLogicalName

        writer = GXXmlWriter()
        objects = ET.Element("Objects")
        lnVersion = 2
        for it in self:
            if isinstance(it, GXDLMSAssociationLogicalName):
                lnVersion = it.version
                break

        for it in self:
            node = ET.SubElement(
                objects, "GXDLMS" + GXDLMSConverter.objectTypeToString(it.objectType)
            )
            if it.shortName != 0:
                ET.SubElement(node, "SN").text = str(it.shortName)
            ET.SubElement(node, "LN").text = it.logicalName
            if it.version != 0:
                ET.SubElement(node, "Version").text = str(it.version)
            if it.description:
                ET.SubElement(node, "Description").text = it.description

            # Add access rights.
            if lnVersion < 3:
                sb = ""
                pos = 1
                while pos <= it.getAttributeCount():
                    sb += str(int(it.getAccess(pos)))
                    pos += 1
                ET.SubElement(node, "Access").text = sb
                sb = ""
                pos = 1
                while pos <= it.getMethodCount():
                    sb += str(int(it.getMethodAccess(pos)))
                    pos += 1
                ET.SubElement(node, "MethodAccess").text = sb
            else:
                sb = ""
                pos = 1
                while pos <= it.getAttributeCount():
                    # Set highest bit to save integer with two chars.
                    value = int(it.getAccess3(pos))
                    value |= 0x8000
                    sb += "%0.2X" % value
                    pos += 1
                ET.SubElement(node, "Access3").text = sb
                sb = ""
                pos = 1
                while pos <= it.getMethodCount():
                    # Set highest bit to save integer with two chars.
                    value = 0x8000 | it.getMethodAccess3(pos)
                    sb += "%0.2X" % value
                    pos += 1
                ET.SubElement(node, "MethodAccess3").text = sb

            if not settings or settings.values:
                writer.objects = []
                writer.objects.append(node)
                it.save(writer)
        str_ = minidom.parseString(
            ET.tostring(objects, encoding="utf-8", method="xml")
        ).toprettyxml(indent="  ")
        with open(name, "w") as f:
            f.write(str_)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXArray import GXArray
from .GXStructure import GXStructure
from .ActionRequestType import ActionRequestType
from .ActionResponseType import ActionResponseType
from .ConfirmedServiceError import ConfirmedServiceError
from .ConnectionState import ConnectionState
from .GetCommandType import GetCommandType
from ._GXAPDU import _GXAPDU
from .GXBitString import GXBitString
from .GXByteBuffer import GXByteBuffer
from .GXTimeZone import GXTimeZone
from .GXDate import GXDate
from .GXDateTime import GXDateTime
from .GXDLMS import GXDLMS
from .GXDLMSAccessItem import GXDLMSAccessItem
from .GXDLMSClient import GXDLMSClient
from .GXDLMSConfirmedServiceError import GXDLMSConfirmedServiceError
from .GXDLMSExceptionResponse import GXDLMSExceptionResponse
from .GXDLMSConnectionEventArgs import GXDLMSConnectionEventArgs
from .GXDLMSConverter import GXDLMSConverter
from .GXDLMSException import GXDLMSException
from .GXDLMSGateway import GXDLMSGateway
from .GXDLMSLimits import GXDLMSLimits
from .GXHdlcSettings import GXHdlcSettings
from .GXDLMSLNCommandHandler import GXDLMSLNCommandHandler
from .GXDLMSLNParameters import GXDLMSLNParameters
from .GXDLMSLongTransaction import GXDLMSLongTransaction
from .GXDLMSNotify import GXDLMSNotify
from .GXDLMSServer import GXDLMSServer
from .GXDLMSSettings import GXDLMSSettings
from .GXDLMSSNCommandHandler import GXDLMSSNCommandHandler
from .GXDLMSSNParameters import GXDLMSSNParameters
from .GXDLMSTranslator import GXDLMSTranslator
from .GXDLMSTranslatorStructure import GXDLMSTranslatorStructure
from .GXDLMSXmlClient import GXDLMSXmlClient
from .GXDLMSXmlPdu import GXDLMSXmlPdu
from .GXDLMSXmlSettings import GXDLMSXmlSettings
from .GXICipher import GXICipher
from .GXReplyData import GXReplyData
from .GXServerReply import GXServerReply
from .GXSNInfo import GXSNInfo
from .GXStandardObisCode import GXStandardObisCode
from .GXStandardObisCodeCollection import GXStandardObisCodeCollection
from .GXTime import GXTime
from .GXWriteItem import GXWriteItem
from .GXXmlLoadSettings import GXXmlLoadSettings
from .HdlcControlFrame import HdlcControlFrame
from ._HDLCInfo import _HDLCInfo
from .MBusCommand import MBusCommand
from .MBusControlInfo import MBusControlInfo
from .MBusEncryptionMode import MBusEncryptionMode
from .MBusMeterType import MBusMeterType
from .ReleaseRequestReason import ReleaseRequestReason
from .ReleaseResponseReason import ReleaseResponseReason
from .SerialnumberCounter import SerialNumberCounter
from .ServiceError import ServiceError
from .SetRequestType import SetRequestType
from .SetResponseType import SetResponseType
from .SingleReadResponse import SingleReadResponse
from .SingleWriteResponse import SingleWriteResponse
from .TranslatorGeneralTags import TranslatorGeneralTags
from .TranslatorSimpleTags import TranslatorSimpleTags
from .TranslatorStandardTags import TranslatorStandardTags
from .TranslatorTags import TranslatorTags
from .ValueEventArgs import ValueEventArgs
from .VariableAccessSpecification import VariableAccessSpecification
from ._GXObjectFactory import _GXObjectFactory
from ._GXFCS16 import _GXFCS16
from .CountType import CountType
from .AesGcmParameter import AesGcmParameter
from .GXCiphering import GXCiphering
from .GXDLMSChippering import GXDLMSChippering
from .GXDLMSChipperingStream import GXDLMSChipperingStream
from .GXEnum import GXEnum
from .GXInt8 import GXInt8
from .GXInt16 import GXInt16
from .GXInt32 import GXInt32
from .GXInt64 import GXInt64
from .GXUInt8 import GXUInt8
from .GXUInt16 import GXUInt16
from .GXUInt32 import GXUInt32
from .GXUInt64 import GXUInt64
from .GXFloat32 import GXFloat32
from .GXFloat64 import GXFloat64
from .GXIntEnum import GXIntEnum
from .GXIntFlag import GXIntFlag
from .GXDLMSTranslatorMessage import GXDLMSTranslatorMessage
from .IGXCryptoNotifier import IGXCryptoNotifier
from .GXCryptoKeyParameter import GXCryptoKeyParameter
name = "gurux_dlms"



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .GXDLMSTarget import GXDLMSTarget


# pylint: disable=too-many-instance-attributes
class GXDLMSParameterMonitor(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSParameterMonitor
    """

    def __init__(self, ln="0.0.16.2.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        # pylint: disable=super-with-arguments
        super(GXDLMSParameterMonitor, self).__init__(
            ObjectType.PARAMETER_MONITOR, ln, sn
        )
        self.parameters = []
        self.captureTime = None
        self.changedParameter = GXDLMSTarget()

    def getValues(self):
        return [
            self.logicalName,
            self.changedParameter,
            self.captureTime,
            self.parameters,
        ]

    #
    # Inserts a new entry in the table.
    #
    # @param client
    # DLMS Client.
    # @param entry
    # Removed entry.
    # If a special day with the same index or with the same date
    #      as an
    # already defined day is inserted, the old entry will be
    # overwritten.
    #
    def insert(self, client, entry):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(3)
        _GXCommon.setData(None, bb, DataType.UINT16, entry.target.objectType)
        _GXCommon.setData(
            None,
            bb,
            DataType.OCTET_STRING,
            _GXCommon.logicalNameToBytes(entry.target.logicalName),
        )
        _GXCommon.setData(None, bb, DataType.INT8, entry.attributeIndex)
        return client.method(self, 1, bb.array(), DataType.ARRAY)

    def delete(self, client, entry):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(3)
        _GXCommon.setData(None, bb, DataType.UINT16, entry.target.objectType)
        _GXCommon.setData(
            None,
            bb,
            DataType.OCTET_STRING,
            _GXCommon.logicalNameToBytes(entry.target.logicalName),
        )
        _GXCommon.setData(None, bb, DataType.INT8, entry.attributeIndex)
        return client.method(self, 2, bb.array(), DataType.ARRAY)

    def invoke(self, settings, e):
        if e.index not in (1, 2):
            e.error = ErrorCode.READ_WRITE_DENIED
        else:
            if e.index == 1:
                # pylint: disable=import-outside-toplevel
                from .._GXObjectFactory import _GXObjectFactory

                tmp = e.parameters
                type_ = ObjectType(tmp[0])
                ln = _GXCommon.toLogicalName(tmp[1])
                index = tmp[2]
                for item in self.parameters:
                    if (
                        item.target.objectType == type_
                        and item.target.logicalName == ln
                        and item.attributeIndex == index
                    ):
                        self.parameters.remove(item)
                        break
                it = GXDLMSTarget()
                it.target = settings.objects.findByLN(type_, ln)
                if it.target is None:
                    it.target = _GXObjectFactory.createObject(type_)
                    it.target.logicalName = ln
                it.attributeIndex = index
                self.parameters.append(it)
            elif e.index == 2:
                tmp = e.parameters
                type_ = ObjectType(tmp[0])
                ln = _GXCommon.toLogicalName(tmp[1])
                index = tmp[2]
                for item in self.parameters:
                    if (
                        item.target.objectType == type_
                        and item.target.logicalName == ln
                        and item.attributeIndex == index
                    ):
                        self.parameters.remove(item)
                        break

    def getAttributeIndexToRead(self, all_):
        attributes = []
        if all_ or not self.logicalName:
            attributes.append(1)
        if all_ or self.canRead(2):
            attributes.append(2)
        if all_ or self.canRead(3):
            attributes.append(3)
        if all_ or self.canRead(4):
            attributes.append(4)
        return attributes

    def getAttributeCount(self):
        return 4

    def getMethodCount(self):
        return 2

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.STRUCTURE
        elif index == 3:
            ret = DataType.OCTET_STRING
        elif index == 4:
            ret = DataType.ARRAY
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    def getValue(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            data = GXByteBuffer()
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(4)
            if self.changedParameter is None:
                _GXCommon.setData(settings, data, DataType.UINT16, 0)
                _GXCommon.setData(
                    settings, data, DataType.OCTET_STRING, [0, 0, 0, 0, 0, 0]
                )
                _GXCommon.setData(settings, data, DataType.INT8, 1)
                _GXCommon.setData(settings, data, DataType.NONE, None)
            else:
                _GXCommon.setData(
                    settings,
                    data,
                    DataType.UINT16,
                    self.changedParameter.target.objectType,
                )
                _GXCommon.setData(
                    settings,
                    data,
                    DataType.OCTET_STRING,
                    _GXCommon.logicalNameToBytes(
                        self.changedParameter.target.logicalName
                    ),
                )
                _GXCommon.setData(
                    settings, data, DataType.INT8, self.changedParameter.attributeIndex
                )
                _GXCommon.setData(
                    settings,
                    data,
                    _GXCommon.getDLMSDataType(self.changedParameter.value),
                    self.changedParameter.value,
                )
            ret = data
        elif e.index == 3:
            ret = self.captureTime
        elif e.index == 4:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            if self.parameters is None:
                data.setUInt8(0)
            else:
                data.setUInt8(len(self.parameters))
                for it in self.parameters:
                    data.setUInt8(DataType.STRUCTURE)
                    data.setUInt8(3)
                    _GXCommon.setData(
                        settings, data, DataType.UINT16, it.target.objectType
                    )
                    _GXCommon.setData(
                        settings,
                        data,
                        DataType.OCTET_STRING,
                        _GXCommon.logicalNameToBytes(it.target.logicalName),
                    )
                    _GXCommon.setData(settings, data, DataType.INT8, it.attributeIndex)
            ret = data
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    def setValue(self, settings, e):
        # pylint: disable=import-outside-toplevel
        from .._GXObjectFactory import _GXObjectFactory

        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.changedParameter = GXDLMSTarget()
            if len(e.value) != 4:
                raise Exception("Invalid structure format.")
            type_ = ObjectType(e.value[0])
            ln = _GXCommon.toLogicalName(e.value[1])
            self.changedParameter.target = settings.objects.findByLN(type_, ln)
            if self.changedParameter.target is None:
                self.changedParameter.target = _GXObjectFactory.createObject(type_)
                self.changedParameter.target.logicalName = ln
            self.changedParameter.attributeIndex = e.value[2]
            self.changedParameter.value = e.value[3]
        elif e.index == 3:
            if e.value is None:
                self.captureTime = None
            else:
                if isinstance(e.value, bytearray):
                    self.captureTime = _GXCommon.changeType(
                        settings, e.value, DataType.DATETIME
                    )
                else:
                    self.captureTime = e.value
        elif e.index == 4:
            self.parameters = []
            if e.value:
                for i in e.value:
                    if len(i) != 3:
                        raise Exception("Invalid structure format.")
                    obj = GXDLMSTarget()
                    type_ = ObjectType(i[0])
                    ln = _GXCommon.toLogicalName(i[1])
                    obj.target = settings.objects.findByLN(type_, ln)
                    if obj.target is None:
                        obj.target = _GXObjectFactory.createObject(type_)
                        obj.target.logicalName = ln
                    obj.attributeIndex = i[2]
                    self.parameters.append(obj)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        # pylint: disable=import-outside-toplevel
        from .._GXObjectFactory import _GXObjectFactory

        self.changedParameter = GXDLMSTarget()
        if reader.isStartElement("ChangedParameter", True):
            ot = ObjectType(reader.readElementContentAsInt("Type"))
            ln = reader.readElementContentAsString("LN")
            self.changedParameter.target = reader.objects.findByLN(ot, ln)
            if self.changedParameter.target is None:
                self.changedParameter.target = _GXObjectFactory.createObject(ot)
                self.changedParameter.target.logicalName = ln
            self.changedParameter.attributeIndex = reader.readElementContentAsInt(
                "Index"
            )
            self.changedParameter.value = reader.readElementContentAsObject(
                "Value", None
            )
            reader.readEndElement("ChangedParameter")
        self.captureTime = reader.readElementContentAsDateTime("Time")
        self.parameters = []
        if reader.isStartElement("Parameters", True):
            while reader.isStartElement("Item", True):
                obj = GXDLMSTarget()
                ot = ObjectType(reader.readElementContentAsInt("Type"))
                ln = reader.readElementContentAsString("LN")
                obj.target = reader.objects.findByLN(ot, ln)
                if obj.target is None:
                    obj.target = _GXObjectFactory.createObject(ot)
                    obj.target.logicalName = ln
                obj.attributeIndex = reader.readElementContentAsInt("Index")
                self.parameters.append(obj)
            reader.readEndElement("Parameters")

    def save(self, writer):
        if self.changedParameter and self.changedParameter.target:
            writer.writeStartElement("ChangedParameter")
            writer.writeElementString(
                "Type", int(self.changedParameter.target.objectType)
            )
            writer.writeElementString("LN", self.changedParameter.target.logicalName)
            writer.writeElementString("Index", self.changedParameter.attributeIndex)
            writer.writeElementObject("Value", self.changedParameter.value)
            writer.writeEndElement()
        writer.writeElementString("Time", self.captureTime)
        writer.writeStartElement("Parameters")
        for it in self.parameters:
            writer.writeStartElement("Item")
            writer.writeElementString("Type", int(it.target.objectType))
            writer.writeElementString("LN", it.target.logicalName)
            writer.writeElementString("Index", it.attributeIndex)
            writer.writeEndElement()
        writer.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from abc import abstractmethod
from ..internal._GXCommon import _GXCommon
from ..manufacturersettings import GXAttributeCollection, GXDLMSAttributeSettings
from ..enums.DataType import DataType
from ..enums.AccessMode import AccessMode
from ..enums.MethodAccessMode import MethodAccessMode
from ..enums.AccessMode3 import AccessMode3
from ..enums.MethodAccessMode3 import MethodAccessMode3


#
# pylint: disable=too-many-public-methods,too-many-instance-attributes,useless-object-inheritance
class GXDLMSObject(object):
    #
    # Constructor,
    #
    def __init__(self, ot, ln=None, sn=0):
        # DLMS version number.
        self.version = 0
        self.objectType = ot
        self.shortName = sn
        # Description of COSEM object.
        self.description = ""
        self.attributes = GXAttributeCollection()
        self.methodAttributes = GXAttributeCollection()
        if ln:
            items = ln.split(".")
            if len(items) != 6:
                raise Exception("Invalid Logical Name.")
        # Logical Name of COSEM object.
        self.logicalName = ln
        self.readTimes = dict()

    #
    # Validate logical name.
    # value: Logical Name.
    #
    @classmethod
    def validateLogicalName(cls, value):
        _GXCommon.logicalNameToBytes(value)

    #
    # Is attribute read.  This can be used with static attributes to
    #      make meter
    # reading faster.
    #
    def isRead(self, index):
        if not self.canRead(index):
            return True
        return not self.getLastReadTime(index)

    def canRead(self, index):
        return int(self.getAccess(index)) and int(AccessMode.READ) != 0

    #
    # Returns time when attribute was last time read.  -
    #
    # @param attributeIndex
    # Attribute index.
    # Is attribute read only.
    #
    def getLastReadTime(self, attributeIndex):
        for k, v in self.readTimes.items():
            if k == attributeIndex:
                return v
        return None

    #
    # Set time when attribute was last time read.
    # @param attributeIndex Attribute index.
    # @param tm Read time.
    #
    def setLastReadTime(self, attributeIndex, tm):
        self.readTimes[attributeIndex] = tm

    #
    # Logical or Short Name of DLMS object.
    #
    def __str__(self):
        str_ = None
        if self.shortName != 0:
            str_ = str(self.shortName)
        else:
            str_ = self.logicalName
        if self.description:
            str_ += " " + self.description
        return str_

    #
    # Interface type of the COSEM object.
    #
    def getObjectType(self):
        return self.objectType

    #
    # The base name of the object, if using SN.  When using SN
    #      referencing,
    # retrieves the base name of the DLMS object.  When using LN
    #      referencing,
    # the value is 0.
    #
    # The base name of the object.
    #
    def getShortName(self):
        return self.shortName

    def getName(self):
        if self.shortName != 0:
            return self.shortName
        return self.logicalName

    #
    # Logical or Short Name of DLMS object.
    #
    # Logical or Short Name of DLMS object
    #
    name = property(getName)

    #
    # Object attribute collection.
    #
    def getAttributes(self):
        return self.attributes

    #
    # Object method attribute collection.
    #
    def getMethodAttributes(self):
        return self.methodAttributes

    #
    # Returns is attribute read only.  -
    #
    # @param index
    # Attribute index.
    # Is attribute read only.
    #
    def getAccess(self, index):
        if index == 1:
            return AccessMode.READ
        att = self.attributes.find(index)
        if att is None:
            return AccessMode.READ_WRITE
        return att.access

    #
    # Set attribute access.
    #
    # @param index
    # Attribute index.
    # @param access
    # Attribute access.
    #
    def setAccess(self, index, access):
        att = self.attributes.find(index)
        if att is None:
            att = GXDLMSAttributeSettings(index)
            self.attributes.append(att)
        att.access = access

    #
    # Returns is attribute read only.  -
    #
    # @param index
    # Attribute index.
    # Is attribute read only.
    #
    def getAccess3(self, index):
        if index == 1:
            return AccessMode3.READ
        att = self.attributes.find(index)
        if att is None:
            return AccessMode3.READ_WRITE
        return att.access3

    #
    # Set attribute access.
    #
    # @param index
    # Attribute index.
    # @param access
    # Attribute access.
    #
    def setAccess3(self, index, access):
        att = self.attributes.find(index)
        if att is None:
            att = GXDLMSAttributeSettings(index)
            self.attributes.append(att)
        att.access3 = access

    #
    # Returns amount of methods.
    #
    @abstractmethod
    def getMethodCount(self):
        raise ValueError("getMethodCount")

    #
    # Returns is Method attribute read only.
    #
    # @param index
    # Method Attribute index.
    # Is attribute read only.
    #
    def getMethodAccess(self, index):
        att = self.getMethodAttributes().find(index)
        if att:
            return att.methodAccess
        return MethodAccessMode.ACCESS

    #
    # Set Method attribute access.
    #
    # @param index
    # Method index.
    # @param access
    # Method access mode.
    #
    def setMethodAccess(self, index, access):
        att = self.getMethodAttributes().find(index)
        if att is None:
            att = GXDLMSAttributeSettings(index)
            self.methodAttributes.append(att)
        att.methodAccess = access

        #

    # Returns is Method attribute read only.
    #
    # @param index
    # Method Attribute index.
    # Is attribute read only.
    #
    def getMethodAccess3(self, index):
        att = self.getMethodAttributes().find(index)
        if att:
            return att.methodAccess3
        return MethodAccessMode3.ACCESS

    #
    # Set Method attribute access.
    #
    # @param index
    # Method index.
    # @param access
    # Method access mode.
    #
    def setMethodAccess3(self, index, access):
        att = self.getMethodAttributes().find(index)
        if att is None:
            att = GXDLMSAttributeSettings(index)
            self.methodAttributes.append(att)
        att.methodAccess3 = access

    def getDataType(self, index):
        att = self.attributes.find(index)
        if not att:
            return DataType.NONE
        return att.type_

    def getUIDataType(self, index):
        att = self.attributes.find(index)
        if not att:
            return DataType.NONE
        return att.uiType

    #
    # Amount of attributes.
    #
    @abstractmethod
    def getAttributeCount(self):
        raise ValueError("getAttributeCount")

    #
    # Object values as an array.
    #
    @abstractmethod
    def getValues(self):
        raise ValueError("getValues")

    @abstractmethod
    def getValue(self, settings, e):
        #
        # Get value.
        # settings: DLMS settings.
        # e: Value event parameters.
        #
        raise ValueError("getValue")

    @abstractmethod
    def setValue(self, settings, e):
        #
        # Set value of given attribute.
        #
        # settings: DLMS settings.
        # e: event parameters.
        raise ValueError("setValue")

    #
    # Server calls this invokes method.
    # @param settings DLMS settings.
    # @param e Value event parameters.
    def invoke(self, settings, e):
        # pylint: disable=unused-argument
        raise ValueError("invoke")

    def setDataType(self, index, type_):
        att = self.attributes.find(index)
        if att is None:
            att = GXDLMSAttributeSettings(index)
            self.attributes.append(att)
        att.type_ = type_

    def setUIDataType(self, index, type_):
        att = self.attributes.find(index)
        if att is None:
            att = GXDLMSAttributeSettings(index)
            self.attributes.append(att)
        att.uiType = type_

    def setStatic(self, index, isStatic):
        att = self.attributes.find(index)
        if att is None:
            att = GXDLMSAttributeSettings(index)
            self.attributes.append(att)
        att.static = isStatic

    def getStatic(self, index):
        att = self.attributes.find(index)
        if att is None:
            att = GXDLMSAttributeSettings(index)
            self.attributes.append(att)
        return att.static



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums.MacState import MacState

# pylint: disable=too-many-instance-attributes
class GXDLMSPrimeNbOfdmPlcMacFunctionalParameters(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSPrimeNbOfdmPlcMacFunctionalParameters
    """

    def __init__(self, ln="0.0.28.3.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.PRIME_NB_OFDM_PLC_MAC_FUNCTIONAL_PARAMETERS, ln, sn)
        #LNID allocated to this node at time of its registration.
        self.lnId = 0
        #LSID allocated to this node at the time of its promotion.
        self.lsId = 0
        #SID of the switch node through which this node is connected to the
        #subnetwork.
        self.sId = 0
        #Subnetwork address to which this node is registered.
        self.sna = None
        #Present functional state of the node.
        self.state = MacState.DISCONNECTED
        #The SCP length, in symbols, in present frame.
        self.scpLength = 0
        #Level of this node in subnetwork hierarchy.
        self.nodeHierarchyLevel = 0
        #Number of beacon slots provisioned in present frame structure.
        self.beaconSlotCount = 0
        #Beacon slot in which this device's switch node transmits its
        #beacon.
        self.beaconRxSlot = 0
        #Beacon slot in which this device transmits its beacon.
        self.beaconTxSlot = 0
        #Number of frames between receptions of two successive beacons.
        self.beaconRxFrequency = 0
        #Number of frames between transmissions of two successive beacons.
        self.beaconTxFrequency = 0
        #This attribute defines the capabilities of the node.
        self.capabilities = []

    def getValues(self):
        return [self.logicalName,
                self.lnId,
                self.lsId,
                self.sId,
                self.sna,
                self.state,
                self.scpLength,
                self.nodeHierarchyLevel,
                self.beaconSlotCount,
                self.beaconRxSlot,
                self.beaconTxSlot,
                self.beaconRxFrequency,
                self.beaconTxFrequency,
                self.capabilities]

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        # LnId
        if all_ or self.canRead(2):
            attributes.append(2)
        # LsId
        if all_ or self.canRead(3):
            attributes.append(3)
        # SId
        if all_ or self.canRead(4):
            attributes.append(4)
        # SNa
        if all_ or self.canRead(5):
            attributes.append(5)
        # State
        if all_ or self.canRead(6):
            attributes.append(6)
        # ScpLength
        if all_ or self.canRead(7):
            attributes.append(7)
        # NodeHierarchyLevel
        if all_ or self.canRead(8):
            attributes.append(8)
        # BeaconSlotCount
        if all_ or self.canRead(9):
            attributes.append(9)
        # BeaconRxSlot
        if all_ or self.canRead(10):
            attributes.append(10)
        # BeaconTxSlot
        if all_ or self.canRead(11):
            attributes.append(11)
        # BeaconRxFrequency
        if all_ or self.canRead(12):
            attributes.append(12)
        # BeaconTxFrequency
        if all_ or self.canRead(13):
            attributes.append(13)
        # Capabilities
        if all_ or self.canRead(14):
            attributes.append(14)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 14

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.INT16
        elif index in (3, 4):
            ret = DataType.UINT8
        elif index == 5:
            ret = DataType.OCTET_STRING
        elif index == 6:
            ret = DataType.ENUM
        elif index == 7:
            ret = DataType.INT16
        elif index in (8, 9, 10, 11, 12, 13):
            ret = DataType.UINT8
        elif index == 14:
            ret = DataType.UINT16
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.lnId
        elif e.index == 3:
            ret = self.lsId
        elif e.index == 4:
            ret = self.sId
        elif e.index == 5:
            ret = self.sna
        elif e.index == 6:
            ret = self.state
        elif e.index == 7:
            ret = self.scpLength
        elif e.index == 8:
            ret = self.nodeHierarchyLevel
        elif e.index == 9:
            ret = self.beaconSlotCount
        elif e.index == 10:
            ret = self.beaconRxSlot
        elif e.index == 11:
            ret = self.beaconTxSlot
        elif e.index == 12:
            ret = self.beaconRxFrequency
        elif e.index == 13:
            ret = self.beaconTxFrequency
        elif e.index == 14:
            ret = self.capabilities
        else:
            ret = None
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.lnId = e.value
        elif e.index == 3:
            self.lsId = e.value
        elif e.index == 4:
            self.sId = e.value
        elif e.index == 5:
            self.sna = e.value
        elif e.index == 6:
            self.state = e.value
        elif e.index == 7:
            self.scpLength = e.value
        elif e.index == 8:
            self.nodeHierarchyLevel = e.value
        elif e.index == 9:
            self.beaconSlotCount = e.value
        elif e.index == 10:
            self.beaconRxSlot = e.value
        elif e.index == 11:
            self.beaconTxSlot = e.value
        elif e.index == 12:
            self.beaconRxFrequency = e.value
        elif e.index == 13:
            self.beaconTxFrequency = e.value
        elif e.index == 14:
            self.capabilities = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.lnId = reader.readElementContentAsInt("LnId")
        self.lsId = reader.readElementContentAsInt("LsId")
        self.sId = reader.readElementContentAsInt("SId")
        self.sna = GXByteBuffer.hexToBytes(reader.readElementContentAsString("SNa"))
        self.state = reader.readElementContentAsInt("State")
        self.scpLength = reader.readElementContentAsInt("ScpLength")
        self.nodeHierarchyLevel = reader.readElementContentAsInt("NodeHierarchyLevel")
        self.beaconSlotCount = reader.readElementContentAsInt("BeaconSlotCount")
        self.beaconRxSlot = reader.readElementContentAsInt("BeaconRxSlot")
        self.beaconTxSlot = reader.readElementContentAsInt("BeaconTxSlot")
        self.beaconRxFrequency = reader.readElementContentAsInt("BeaconRxFrequency")
        self.beaconTxFrequency = reader.readElementContentAsInt("BeaconTxFrequency")
        self.capabilities = reader.readElementContentAsInt("Capabilities")

    def save(self, writer):
        writer.writeElementString("LnId", self.lnId)
        writer.writeElementString("LsId", self.lsId)
        writer.writeElementString("SId", self.sId)
        writer.writeElementString("SNa", self.sna)
        writer.writeElementString("State", self.state)
        writer.writeElementString("ScpLength", self.scpLength)
        writer.writeElementString("NodeHierarchyLevel", self.nodeHierarchyLevel)
        writer.writeElementString("BeaconSlotCount", self.beaconSlotCount)
        writer.writeElementString("BeaconRxSlot", self.beaconRxSlot)
        writer.writeElementString("BeaconTxSlot", self.beaconTxSlot)
        writer.writeElementString("BeaconRxFrequency", self.beaconRxFrequency)
        writer.writeElementString("BeaconTxFrequency", self.beaconTxFrequency)
        writer.writeElementString("Capabilities", self.capabilities)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..GXDateTime import GXDateTime
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .GXDLMSSeasonProfile import GXDLMSSeasonProfile
from .GXDLMSWeekProfile import GXDLMSWeekProfile
from .GXDLMSDayProfile import GXDLMSDayProfile
from .GXDLMSDayProfileAction import GXDLMSDayProfileAction

# pylint: disable=too-many-public-methods, too-many-instance-attributes
class GXDLMSActivityCalendar(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSActivityCalendar
    """

    def __init__(self, ln="0.0.13.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.ACTIVITY_CALENDAR, ln, sn)
        self.calendarNameActive = None
        self.seasonProfileActive = None
        self.weekProfileTableActive = None
        self.dayProfileTableActive = None
        self.calendarNamePassive = None
        self.seasonProfilePassive = None
        self.weekProfileTablePassive = None
        self.dayProfileTablePassive = None
        self.time = None
        self.isSec = False

    def getValues(self):
        return [self.logicalName,
                self.calendarNameActive,
                self.seasonProfileActive,
                self.weekProfileTableActive,
                self.dayProfileTableActive,
                self.calendarNamePassive,
                self.seasonProfilePassive,
                self.weekProfileTablePassive,
                self.dayProfileTablePassive,
                self.time]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  CalendarNameActive
        if all_ or self.canRead(2):
            attributes.append(2)
        #  SeasonProfileActive
        if all_ or self.canRead(3):
            attributes.append(3)
        #  WeekProfileTableActive
        if all_ or self.canRead(4):
            attributes.append(4)
        #  DayProfileTableActive
        if all_ or self.canRead(5):
            attributes.append(5)
        #  CalendarNamePassive
        if all_ or self.canRead(6):
            attributes.append(6)
        #  SeasonProfilePassive
        if all_ or self.canRead(7):
            attributes.append(7)
        #  WeekProfileTablePassive
        if all_ or self.canRead(8):
            attributes.append(8)
        #  DayProfileTablePassive
        if all_ or self.canRead(9):
            attributes.append(9)
        #  Time.
        if all_ or self.canRead(10):
            attributes.append(10)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 10

    #
    # This method copies all passive properties to the active.
    #
    def activatePassiveCalendar(self, client):
        return client.method(self, 1, int(0), DataType.INT8)

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 1

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.OCTET_STRING
        elif index == 3:
            ret = DataType.ARRAY
        elif index == 4:
            ret = DataType.ARRAY
        elif index == 5:
            ret = DataType.ARRAY
        elif index == 6:
            ret = DataType.OCTET_STRING
        elif index == 7:
            ret = DataType.ARRAY
        elif index == 8:
            ret = DataType.ARRAY
        elif index == 9:
            ret = DataType.ARRAY
        elif index == 10:
            if self.isSec:
                ret = DataType.DATETIME
            else:
                ret = DataType.OCTET_STRING
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    @classmethod
    def __getSeasonProfile(cls, settings, target):
        data = GXByteBuffer()
        data.setUInt8(DataType.ARRAY)
        if target is None:
            #  Add count
            _GXCommon.setObjectCount(0, data)
        else:
            #  Add count
            _GXCommon.setObjectCount(len(target), data)
            for it in target:
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(3)
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, it.name)
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, it.start)
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, it.weekName)
        return data.array()

    @classmethod
    def __getWeekProfileTable(cls, settings, target):
        data = GXByteBuffer()
        data.setUInt8(DataType.ARRAY)
        if target is None:
            #  Add count
            _GXCommon.setObjectCount(0, data)
        else:
            #  Add count
            _GXCommon.setObjectCount(len(target), data)
            for it in target:
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(8)
                _GXCommon.setData(settings, data, DataType.OCTET_STRING, it.name)
                _GXCommon.setData(settings, data, DataType.UINT8, it.monday)
                _GXCommon.setData(settings, data, DataType.UINT8, it.tuesday)
                _GXCommon.setData(settings, data, DataType.UINT8, it.wednesday)
                _GXCommon.setData(settings, data, DataType.UINT8, it.thursday)
                _GXCommon.setData(settings, data, DataType.UINT8, it.friday)
                _GXCommon.setData(settings, data, DataType.UINT8, it.saturday)
                _GXCommon.setData(settings, data, DataType.UINT8, it.sunday)
        return data.array()

    @classmethod
    def __getDayProfileTable(cls, settings, target):
        data = GXByteBuffer()
        data.setUInt8(DataType.ARRAY)
        if target is None:
            #  Add count
            _GXCommon.setObjectCount(0, data)
        else:
            #  Add count
            _GXCommon.setObjectCount(len(target), data)
            for it in target:
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(2)
                _GXCommon.setData(settings, data, DataType.UINT8, it.dayId)
                data.setUInt8(DataType.ARRAY)
                #  Add count
                _GXCommon.setObjectCount(len(it.daySchedules), data)
                for action in it.daySchedules:
                    data.setUInt8(DataType.STRUCTURE)
                    data.setUInt8(3)
                    _GXCommon.setData(settings, data, DataType.OCTET_STRING, action.startTime)
                    _GXCommon.setData(settings, data, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(action.scriptLogicalName))
                    _GXCommon.setData(settings, data, DataType.UINT16, int(action.scriptSelector))
        return data.array()

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            #pylint: disable=bad-option-value,redefined-variable-type
            if not self.calendarNameActive:
                ret = None
            elif self.isSec:
                ret = GXByteBuffer.hexToBytes(self.calendarNameActive)
            else:
                ret = self.calendarNameActive.encode("utf-8")
        elif e.index == 3:
            ret = self.__getSeasonProfile(settings, self.seasonProfileActive)
        elif e.index == 4:
            ret = self.__getWeekProfileTable(settings, self.weekProfileTableActive)
        elif e.index == 5:
            ret = self.__getDayProfileTable(settings, self.dayProfileTableActive)
        elif e.index == 6:
            if self.calendarNamePassive is None:
                ret = None
            elif self.isSec:
                ret = GXByteBuffer.hexToBytes(self.calendarNamePassive)
            else:
                ret = self.calendarNamePassive.encode()
        elif e.index == 7:
            ret = self.__getSeasonProfile(settings, self.seasonProfilePassive)
        elif e.index == 8:
            ret = self.__getWeekProfileTable(settings, self.weekProfileTablePassive)
        elif e.index == 9:
            ret = self.__getDayProfileTable(settings, self.dayProfileTablePassive)
        elif e.index == 10:
            ret = self.time
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    @classmethod
    def __setSeasonProfile(cls, settings, value):
        items = []
        if value:
            for item in value:
                it = GXDLMSSeasonProfile()
                it.name = item[0]
                it.start = _GXCommon.changeType(settings, item[1], DataType.DATETIME)
                it.weekName = item[2]
                items.append(it)
        return items

    @classmethod
    def __setWeekProfileTable(cls, value):
        items = []
        if value:
            for item in value:
                it = GXDLMSWeekProfile()
                it.name = item[0]
                it.monday = item[1]
                it.tuesday = item[2]
                it.wednesday = item[3]
                it.thursday = item[4]
                it.friday = item[5]
                it.saturday = item[6]
                it.sunday = item[7]
                items.append(it)
        return items

    @classmethod
    def __setDayProfileTable(cls, settings, value):
        items = []
        if value:
            for item in value:
                it = GXDLMSDayProfile()
                it.dayId = item[0]
                it.daySchedules = []
                for it2 in item[1]:
                    ac = GXDLMSDayProfileAction()
                    ac.startTime = _GXCommon.changeType(settings, it2[0], DataType.TIME)
                    ac.scriptLogicalName = _GXCommon.toLogicalName(it2[1])
                    ac.scriptSelector = it2[2]
                    it.daySchedules.append(ac)
                items.append(it)
        return items

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if self.isSec or not GXByteBuffer.isAsciiString(e.value):
                self.calendarNameActive = GXByteBuffer.hex(e.value)
            else:
                self.calendarNameActive = e.value.decode("utf-8").strip('\x00')
        elif e.index == 3:
            self.seasonProfileActive = self.__setSeasonProfile(settings, e.value)
        elif e.index == 4:
            self.weekProfileTableActive = self.__setWeekProfileTable(e.value)
        elif e.index == 5:
            self.dayProfileTableActive = self.__setDayProfileTable(settings, e.value)
        elif e.index == 6:
            if self.isSec or not GXByteBuffer.isAsciiString(e.value):
                self.calendarNamePassive = GXByteBuffer.hex(e.value)
            else:
                self.calendarNamePassive = e.value.decode("utf-8").strip('\x00')
        elif e.index == 7:
            self.seasonProfilePassive = self.__setSeasonProfile(settings, e.value)
        elif e.index == 8:
            self.weekProfileTablePassive = self.__setWeekProfileTable(e.value)
        elif e.index == 9:
            self.dayProfileTablePassive = self.__setDayProfileTable(settings, e.value)
        elif e.index == 10:
            if isinstance(e.value, GXDateTime):
                self.time = e.value
            else:
                self.time = _GXCommon.changeType(settings, e.value, DataType.DATETIME)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    @classmethod
    def loadSeasonProfile(cls, reader, name):
        list_ = []
        if reader.isStartElement(name, True):
            while reader.isStartElement("Item", True):
                it = GXDLMSSeasonProfile()
                it.name = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Name"))
                it.start = reader.readElementContentAsDateTime("Start")
                it.weekName = GXByteBuffer.hexToBytes(reader.readElementContentAsString("WeekName"))
                list_.append(it)
            reader.readEndElement(name)
        return list_

    @classmethod
    def loadWeekProfileTable(cls, reader, name):
        list_ = []
        if reader.isStartElement(name, True):
            while reader.isStartElement("Item", True):
                it = GXDLMSWeekProfile()
                it.name = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Name"))
                it.monday = reader.readElementContentAsInt("Monday")
                it.tuesday = reader.readElementContentAsInt("Tuesday")
                it.wednesday = reader.readElementContentAsInt("Wednesday")
                it.thursday = reader.readElementContentAsInt("Thursday")
                it.friday = reader.readElementContentAsInt("Friday")
                it.saturday = reader.readElementContentAsInt("Saturday")
                it.sunday = reader.readElementContentAsInt("Sunday")
                list_.append(it)
            reader.readEndElement(name)
        return list_

    @classmethod
    def loadDayProfileTable(cls, reader, name):
        list_ = []
        if reader.isStartElement(name, True):
            while reader.isStartElement("Item", True):
                it = GXDLMSDayProfile()
                it.dayId = reader.readElementContentAsInt("DayId")
                list_.append(it)
                it.daySchedules = []
                if reader.isStartElement("Actions", True):
                    while reader.isStartElement("Action", True):
                        d = GXDLMSDayProfileAction()
                        it.daySchedules.append(d)
                        d.startTime = reader.readElementContentAsTime("Start")
                        d.scriptLogicalName = reader.readElementContentAsString("LN")
                        d.scriptSelector = reader.readElementContentAsInt("Selector")
                    reader.readEndElement("Actions")
            reader.readEndElement(name)
        return list_

    def load(self, reader):
        self.calendarNameActive = reader.readElementContentAsString("CalendarNameActive")
        self.seasonProfileActive = self.loadSeasonProfile(reader, "SeasonProfileActive")
        self.weekProfileTableActive = self.loadWeekProfileTable(reader, "WeekProfileTableActive")
        self.dayProfileTableActive = self.loadDayProfileTable(reader, "DayProfileTableActive")
        self.calendarNamePassive = reader.readElementContentAsString("CalendarNamePassive")
        self.seasonProfilePassive = self.loadSeasonProfile(reader, "SeasonProfilePassive")
        self.weekProfileTablePassive = self.loadWeekProfileTable(reader, "WeekProfileTablePassive")
        self.dayProfileTablePassive = self.loadDayProfileTable(reader, "DayProfileTablePassive")
        self.time = reader.readElementContentAsDateTime("Time")

    @classmethod
    def saveSeasonProfile(cls, writer, list_, name):
        if list_:
            writer.writeStartElement(name)
            for it in list_:
                writer.writeStartElement("Item")
                str_ = it.name
                if isinstance(str_, (bytearray, bytes)):
                    str_ = GXByteBuffer.hex(str_)
                writer.writeElementString("Name", str_)
                writer.writeElementString("Start", it.start)
                writer.writeElementString("WeekName", GXByteBuffer.hex(it.weekName))
                writer.writeEndElement()
            writer.writeEndElement()

    @classmethod
    def saveWeekProfileTable(cls, writer, list_, name):
        if list_:
            writer.writeStartElement(name)
            for it in list_:
                writer.writeStartElement("Item")
                writer.writeElementString("Name", GXByteBuffer.hex(it.name))
                writer.writeElementString("Monday", it.monday)
                writer.writeElementString("Tuesday", it.tuesday)
                writer.writeElementString("Wednesday", it.wednesday)
                writer.writeElementString("Thursday", it.thursday)
                writer.writeElementString("Friday", it.friday)
                writer.writeElementString("Saturday", it.saturday)
                writer.writeElementString("Sunday", it.sunday)
                writer.writeEndElement()
            writer.writeEndElement()

    @classmethod
    def saveDayProfileTable(cls, writer, list_, name):
        if list_:
            writer.writeStartElement(name)
            for it in list_:
                writer.writeStartElement("Item")
                writer.writeElementString("DayId", it.dayId)
                writer.writeStartElement("Actions")
                for d in it.daySchedules:
                    writer.writeStartElement("Action")
                    writer.writeElementString("Start", d.startTime)
                    writer.writeElementString("LN", d.scriptLogicalName)
                    writer.writeElementString("Selector", d.scriptSelector)
                    writer.writeEndElement()
                writer.writeEndElement()
                writer.writeEndElement()
            writer.writeEndElement()

    def save(self, writer):
        writer.writeElementString("CalendarNameActive", self.calendarNameActive)
        self.saveSeasonProfile(writer, self.seasonProfileActive, "SeasonProfileActive")
        self.saveWeekProfileTable(writer, self.weekProfileTableActive, "WeekProfileTableActive")
        self.saveDayProfileTable(writer, self.dayProfileTableActive, "DayProfileTableActive")
        writer.writeElementString("CalendarNamePassive", self.calendarNamePassive)
        self.saveSeasonProfile(writer, self.seasonProfilePassive, "SeasonProfilePassive")
        self.saveWeekProfileTable(writer, self.weekProfileTablePassive, "WeekProfileTablePassive")
        self.saveDayProfileTable(writer, self.dayProfileTablePassive, "DayProfileTablePassive")
        writer.writeElementString("Time", self.time)
